# -*- coding: utf-8 -*-
#
# This file is part of the E2RingSimulator project
#
# Copyright (C): 2019
#                European Synchrotron Radiation Facility
#                BP 220, Grenoble 38043
#                France
#
# Distributed under the terms of the LGPL license.
# See LICENSE.txt for more info.

"""
RingSimulator

This class is the interface to the Elettra 2.0 storage ring simulator.
Simulation is done using the python Accelerator Toolbox (pyAT) 
package.

Modified and adapted from the ESRF EBS storage ring simulator
"""

# PROTECTED REGION ID(E2RingSimulator.system_imports) ENABLED START #
# PROTECTED REGION END #    //  E2RingSimulator.system_imports

# PyTango imports
import tango
from tango import DebugIt
from tango.server import run
from tango.server import Device
from tango.server import attribute, command
from tango.server import device_property
from tango import AttrQuality, DispLevel, DevState
from tango import AttrWriteType, PipeWriteType
import enum
# Additional import
# PROTECTED REGION ID(E2RingSimulator.additionnal_import) ENABLED START #
import at
import TangoNameGenerator as tgn
import E2RingThread as e2th
import E2RingUpdateThread as e2update
from ElementProxy import *
import threading
import numpy as np


# PROTECTED REGION END #    //  E2RingSimulator.additionnal_import

__all__ = ["E2RingSimulator", "main"]


class Mode(enum.IntEnum):
    """Python enumerated type for Mode attribute."""
    TbT = 0
    Atlinopt = 1
    Atx = 2
    Mixed = 3


class E2RingSimulator(Device):
    """
    This class is the interface to the Elettra 2.0 storage ring simulator.
    Simulation is done using the python Accelerator Toolbox (pyAT) 
    package.
    
    Modified and adapted from the ESRF EBS storage ring simulator

    **Properties:**

    - Device Property
        RingFile
            - The ring description file
            - Type:'str'
        Verbose
            - Verbose mode for the simulation loop
            - Type:'bool'
    """
    # PROTECTED REGION ID(E2RingSimulator.class_variable) ENABLED START #
    TBT_COORD_NB = 6
    # PROTECTED REGION END #    //  E2RingSimulator.class_variable

    # -----------------
    # Device Properties
    # -----------------

    RingFile = device_property(
        dtype='str',
        doc="The ring description file",
        mandatory=True
    )

    Verbose = device_property(
        dtype='bool',
        default_value=False,
        doc="Verbose mode for the simulation loop"
    )

    # ----------
    # Attributes
    # ----------

    Tune_h = attribute(
        dtype='DevDouble',
        label="Horizontal tune",
    )

    Tune_v = attribute(
        dtype='DevDouble',
        label="Vertical tune",
    )

    Chromaticity_h = attribute(
        dtype='DevDouble',
        label="Horizontal chromaticity",
    )

    Chromaticity_v = attribute(
        dtype='DevDouble',
        label="Vertical chromaticity",
    )

    Emittance_h = attribute(
        dtype='DevDouble',
        label="Horizontal emittance",
        unit="pm",
        standard_unit="1",
        display_unit="1e12",
        doc="Horizontal emittance",
    )

    Emittance_v = attribute(
        dtype='DevDouble',
        label="Vertical emittance",
        unit="pm",
        standard_unit="1.0",
        display_unit="1e12",
        doc="Vertical emittance",
    )

    RfFrequency = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="Rf Frequency",
        unit="MHz",
        standard_unit="1.0",
        display_unit="1e-06",
        format="%12.8f",
        max_value=359000000,
    )

    RfVoltage = attribute(
        dtype='DevDouble',
        access=AttrWriteType.READ_WRITE,
        label="Rf Voltage",
        unit="MV",
        standard_unit="1.0",
        display_unit="1e-06",
        format="%8.6f",
        max_value=9000000,
    )

    Radiation = attribute(
        dtype='DevBoolean',
        access=AttrWriteType.READ_WRITE,
    )

    LoopTime = attribute(
        dtype='DevDouble',
        unit="s",
        standard_unit="1",
        display_unit="1",
        doc="Py AT loop time (in sec)",
    )

    Counter = attribute(
        dtype='DevLong64',
        display_level=DispLevel.EXPERT,
        label="SimuLoopCounter",
        doc="Simulator loop counter",
    )

    Mode = attribute(
        dtype=Mode,
        access=AttrWriteType.READ_WRITE,
        display_level=DispLevel.EXPERT,
        doc="Simulator mode",
    )

    AtxEvery = attribute(
        dtype='DevLong',
        access=AttrWriteType.READ_WRITE,
        display_level=DispLevel.EXPERT,
        unit="loop",
        max_value=100,
        min_value=1,
        memorized=True,
        hw_memorized=True,
        doc="n mixed mode, one call to atx every XX loops",
    )

    RingName = attribute(
        dtype='DevString',
        access=AttrWriteType.READ_WRITE,
        display_level=DispLevel.EXPERT,
        doc="Ring name from Py AT",
    )

    SimulationStarted = attribute(
        dtype='DevBoolean',
        display_level=DispLevel.EXPERT,
        doc="Set to true when the simulation loop is MOVING. Change event is fired on this attribute",
    )

    TbT_BufferSize = attribute(
        dtype='DevLong',
        access=AttrWriteType.READ_WRITE,
        display_level=DispLevel.EXPERT,
        unit="turn",
        max_value=5000,
        min_value=-5000,
        doc="Buffer size (in turn number) for the Turn By Turn mode",
    )

    ErrMessage = attribute(
        dtype='DevString',
        access=AttrWriteType.WRITE,
        display_level=DispLevel.EXPERT,
    )

    Energy = attribute(
        dtype='DevDouble',
        label="nominal ring energy",
        unit="eV",
        standard_unit="1",
        display_unit="eV",
        format="%2.6e",
        doc="Nominal ring energy [eV]",
    )

    UpdateTime = attribute(
        dtype='DevDouble',
        unit="s",
        standard_unit="1",
        display_unit="s",
        doc="devices read and update time (in sec)",
    )

    ClosedOrbit_h = attribute(
        dtype=('DevDouble',),
        max_dim_x=320,
        label="Horizontal closed orbit",
        unit="m",
    )

    ClosedOrbit_v = attribute(
        dtype=('DevDouble',),
        max_dim_x=320,
        label="Vertical closed orbit",
        unit="m",
    )

    Beta_h = attribute(
        dtype=('DevDouble',),
        max_dim_x=320,
        label="Horizontal beta function",
        unit="m",
    )

    Beta_v = attribute(
        dtype=('DevDouble',),
        max_dim_x=320,
        label="Vertical beta function",
        unit="m",
    )

    Eta_h = attribute(
        dtype=('DevDouble',),
        max_dim_x=320,
        label="Horizontal dispersion function",
        unit="m",
    )

    Eta_v = attribute(
        dtype=('DevDouble',),
        max_dim_x=320,
        label="Vertical dispersion function",
        unit="m",
    )

    TbT_InCoord = attribute(
        dtype=('DevDouble',),
        access=AttrWriteType.READ_WRITE,
        max_dim_x=6,
        display_level=DispLevel.EXPERT,
        label="TbTInputCoordinates",
        doc="arr[0] = x, arr[1] = x`, arr[2] = z, arr[3] = z`. arr[4] = dp/p, arr[5] = ct",
    )

    Emittances_BPM_H = attribute(
        dtype=('DevDouble',),
        max_dim_x=320,
    )

    Emittances_BPM_V = attribute(
        dtype=('DevDouble',),
        max_dim_x=320,
    )

    Emittances_PinHoles_H = attribute(
        dtype=('DevDouble',),
        max_dim_x=5,
    )

    Emittances_PinHoles_V = attribute(
        dtype=('DevDouble',),
        max_dim_x=5,
    )

    S_bpms = attribute(
        dtype=('DevDouble',),
        access=AttrWriteType.READ_WRITE,
        max_dim_x=320,
    )

    S_pinholes = attribute(
        dtype=('DevDouble',),
        access=AttrWriteType.READ_WRITE,
        max_dim_x=5,
    )

    HPositionsTbT = attribute(
        dtype=(('DevDouble',),),
        max_dim_x=5000, max_dim_y=320,
        unit="mm",
        standard_unit="1.0",
        display_unit="1000.0",
        format="%6.2f",
        doc="H positions in TbT mode. In the iimage, the X dimension is the turn number while the \nY dimension is the BPM number",
    )

    VPositionsTbT = attribute(
        dtype=(('DevDouble',),),
        max_dim_x=5000, max_dim_y=320,
        unit="mm",
        standard_unit="1.0",
        display_unit="1000.0",
        format="%6.2f",
        doc="V positions in TbT mode. In the image, the X dimension is the turn number while the \nY dimension is the BPM number",
    )

    # ---------------
    # General methods
    # ---------------

    def init_device(self):
        """Initializes the attributes and properties of the E2RingSimulator."""
        Device.init_device(self)
        self.Mode.set_data_ready_event(True)
        self.set_change_event("SimulationStarted", True, False)
        self._tune_h = 0.0
        self._tune_v = 0.0
        self._chromaticity_h = 0.0
        self._chromaticity_v = 0.0
        self._emittance_h = 0.0
        self._emittance_v = 0.0
        self._rf_frequency = 0.0
        self._rf_voltage = 0.0
        self._radiation = False
        self._loop_time = 0.0
        self._counter = 0
        self._mode = Mode.TbT
        self._atx_every = 0
        self._ring_name = ''
        self._simulation_started = False
        self._tb_t__buffer_size = 0
        self._err_message = ''
        self._energy = 0.0
        self._update_time = 0.0
        self._closed_orbit_h = (0.0,)
        self._closed_orbit_v = (0.0,)
        self._beta_h = (0.0,)
        self._beta_v = (0.0,)
        self._eta_h = (0.0,)
        self._eta_v = (0.0,)
        self._tb_t__in_coord = (0.0,)
        self._emittances__bpm__h = (0.0,)
        self._emittances__bpm__v = (0.0,)
        self._emittances__pin_holes__h = (0.0,)
        self._emittances__pin_holes__v = (0.0,)
        self._s_bpms = (0.0,)
        self._s_pinholes = (0.0,)
        self._h_positions_tb_t = ((0.0,),)
        self._v_positions_tb_t = ((0.0,),)
        # PROTECTED REGION ID(E2RingSimulator.init_device) ENABLED START #
            
        self.thread = None
        self.update_thread = None
        self.ring = None
        
        
        self.ring = at.load_lattice(self.RingFile)
        if self.ring is None:
            self.set_state(tango.DevState.UNKNOWN)
            self.set_status("failed to load ring file")
            self.__err_message ="failed to load ring file"
            self.fatal_stream("failed to load ring file %s",self.RingFile)
        
            
            return
        self.set_state(tango.DevState.STANDBY)
        self.set_status("standby")
        self._energy = self.ring.energy
        self._ring_name = self.ring.name + " " + self.RingFile
        self._rf_frequency = self.ring.get_rf_frequency()
        self._rf_voltage = self.ring.get_rf_voltage()
        
        # get indexes (once and for all)
        # elems = self.ring.get_elements('BPM*') or self.ring.get_elements(at.element.Monitor)
        # self.ind_bpms = [self.ring.index(elem) for elem in elems]

        self.ind_bpms = list(at.get_refpts(self.ring, 'BPM*'))
        self._num_bpms = len(self.ind_bpms)
        self._s_bpms = self.ring.get_s_pos(refpts=self.ind_bpms)
        self.info_stream("num BPM %d",self._num_bpms)
        
        self.ind_rf = list(at.get_refpts(self.ring, 'CAV*'))
        self.info_stream("RF %d", len(self.ind_rf))
        #self.ind_pinhole = list(get_refpts(self.ring, 'PINHOLE*'))
        self.ind_pinhole = self.ind_rf # trucco per iniziare  far girare il codice - ci servono i pinhole dell anello
        self.ind_bpm_pinhole = self.ind_bpms + self.ind_pinhole
        self.ind_bpm_pinhole.sort()
        self._s_pinholes = self.ring.get_s_pos(refpts=self.ind_pinhole)
        
        # init masks for selecting functions ate measuerement points

        self.mask_bpm = [self.ind_bpm_pinhole.index(x) for x in self.ind_bpms]
        self.mask_pin = [self.ind_bpm_pinhole.index(x) for x in self.ind_pinhole]
       
        #pinhole sigma exchange data
        self.sigmah_pin =  [np.NaN] * len(self.ind_bpm_pinhole)
        self.sigmav_pin =  [np.NaN] * len(self.ind_bpm_pinhole)
        self.sigmahv_pin = [np.NaN] * len(self.ind_bpm_pinhole)
        
        # init some values
        dev_m_attr = self.get_device_attr()
        
        self._tb_t__buffer_size = 1000
        tbt = dev_m_attr.get_w_attr_by_name("TbT_BufferSize")
        tbt.set_write_value(self._tb_t__buffer_size)
        
        self._tb_t__in_coord = np.array([0.0,0.0,0.0,0.0,0.0,0.0])
        tbt = dev_m_attr.get_w_attr_by_name("TbT_InCoord")
        tbt.set_write_value(self._tb_t__in_coord)
        
        self._mode = Mode.Mixed
        md = dev_m_attr.get_w_attr_by_name("Mode")
        md.set_write_value(self._mode)
        
        # start with radiation OFF
        rd = dev_m_attr.get_w_attr_by_name("Radiation")
        self._radiation = False
        rd.set_write_value(self._radiation )
        self.ring.radiation_off()
        # changing radiation interacts with the update thread - use a lock to guard
        self._radlock = threading.Lock()
        
        """
        questo servirà più avanti
        self.ind_id = list(at.get_refpts(self.ring, 'ID*'))
        self.info_stream("ID %d", len(self.ind_id))
        """
        
        # set RF and voltage to model one
        self.ligth_speed = 2.99792458e08
        self.circumference = self.ring.s_range[-1]

        self.harm = self.ring[self.ind_rf[0]].HarmNumber

        self._rf_frequency = self.ring[self.ind_rf[0]].Frequency
        rff = dev_m_attr.get_w_attr_by_name("RfFrequency")
        rff.set_write_value(self._rf_frequency)
        
        self._rf_voltage = sum([self.ring[f].Voltage for f in self.ind_rf])
        rfv = dev_m_attr.get_w_attr_by_name("RfVoltage")
        rfv.set_write_value(self._rf_voltage)
        
        self.info_stream('RF: %f Hz, %f V, %f bunches', self._rf_frequency, self._rf_voltage, self.harm)
        
        self.info_stream('Circumference: %f m',self.circumference)
        
        
         # compute/store some other usefull lattice related quantities
        self.delta_l = ((self.ligth_speed * self.harm) / self._rf_frequency) - self.circumference
        
        #build the dictionaries of elements to be scanned for optics updates
        #build magnet list and syntesize the device list - device names are set as Device attribute in the lattice model -Elettra2 specfic
        tn = tgn.TangoNameGenerator("sr", "magnet", False)
        quad_idx = self.ring.uint32_refpts(at.Quadrupole)

        # quadrupoles
        self.qlist = []
        self.qeleproxy = [] # list of ElementProxy objetcs for quadrupoles
        for iq in quad_idx:
            q=self.ring[iq]
            qdevname = tn.dev_name(q.FamName)
            # print("quad name", qdevname, q.FamName, q.K, q.PolynomB[1])
            
            self.ring[iq].Device = qdevname #device attribyte in at element . Necessario ?
            self.qlist.append(qdevname)
            self.debug_stream("Q: %s", qdevname)
            # qeleproxy = ElementProxy(q.FamName, qdevname, iq, q.PolynomB[1])
            self.qeleproxy.append(ElementProxy(q.FamName, qdevname, iq, q.PolynomB[1]))
            
        # sextupoles
        sext_idx = self.ring.uint32_refpts(at.Sextupole)
        self.slist = []
        self.seleproxy = [] # list of ElementProxy objetcs for sextupoles
        for idsxt in sext_idx:
            s = self.ring[idsxt]
            sexdevname = tn.dev_name(s.FamName)
            self.ring[idsxt].Device = sexdevname
            self.slist.append(sexdevname)
            # print("sext dev name ",sexdevname, hasattr(s, 'SkewID'), s.FamName)
            self.debug_stream("S: %s", sexdevname)
            self.seleproxy.append(ElementProxy(s.FamName, sexdevname, idsxt, s.PolynomB[2]))
            
         # soctupoles
        oct_idx = self.ring.uint32_refpts(at.Octupole)
        self.olist = []
        for idoct in oct_idx:
            o = self.ring[idoct]
            octdevname = tn.dev_name(o.FamName)
            self.ring[idoct].Device = octdevname
            self.olist.append(octdevname)
            self.debug_stream("O: %s", octdevname)
        


        # pure correctors
        corr_idx = self.ring.uint32_refpts(at.Corrector)
        self.clist = []
        self.celeproxy = [] # list of CorrectorProxy objetcs for correctors
        for ic in corr_idx:
            c=self.ring[ic]
            corrdevname = tn.dev_name(c.FamName)
            #element
            self.ring[ic].Device = corrdevname #device attribyte in at element . Necessario ?
            self.clist.append(corrdevname)
            self.debug_stream("C: %s", corrdevname)
            #print(c.FamName, corrdevname, ic)
            celeproxy = CorrectorProxy(c.FamName, corrdevname, ic, 0.0)
            if celeproxy.valid:  # TO BE FIXED: temporary fix of pass method for corrctor - shoul be done in the lattice file
                 self.ring[ic].PassMethod ='CorrectorPass'
                 # print(c.FamName, corrdevname,self.ring[ic].PassMethod )
            self.celeproxy.append(CorrectorProxy(c.FamName, corrdevname, ic, 0.0))
        
       
        
        # for check only prit qeleproxy list - rmove in production
        
        for ep in self.qeleproxy:
            self.debug_stream("%s %s %s", ep.mag_name, ep.dev_name, ep.valid)
        
        for sp in self.seleproxy:
            self.debug_stream("%s %s %s", sp.mag_name, sp.dev_name, sp.valid)
        
        for cp in self.celeproxy:
            self.debug_stream("%s %s %s", cp.mag_name, cp.dev_name, cp.valid)
        
         #create the thread as run it
        self.thread = e2th.E2RingThread(self)
        self.thread.start()
        self.update_thread = e2update.E2RingUpdateThread(self)
        self.update_thread.start()
        # PROTECTED REGION END #    //  E2RingSimulator.init_device

    def always_executed_hook(self):
        """Method always executed before any TANGO command is executed."""
        # PROTECTED REGION ID(E2RingSimulator.always_executed_hook) ENABLED START #
        # PROTECTED REGION END #    //  E2RingSimulator.always_executed_hook

    def delete_device(self):
        """Hook to delete resources allocated in init_device.

        This method allows for any memory or other resources allocated in the
        init_device method to be released.  This method is called by the device
        destructor and by the device Init command.
        """
        # PROTECTED REGION ID(E2RingSimulator.delete_device) ENABLED START #
        if self.thread:
            self.thread.stop()
            self.debug_stream("joining thread")
            try:
                self.thread.join()
            except Exception as e:
                print("The exception has been Handled in the Main, Details of the Exception are:\n",e)
            self.thread = None
        if self.update_thread:
            self.update_thread.stop()
            self.debug_stream("joining thread")
            try:
                self.update_thread.join()
            except Exception as e:
                print("The exception has been Handled in the Main, Details of the Exception are:\n",e)
            self.update_thread = None
        
        # PROTECTED REGION END #    //  E2RingSimulator.delete_device

    # ------------------
    # Attributes methods
    # ------------------

    def read_Tune_h(self):
        # PROTECTED REGION ID(E2RingSimulator.Tune_h_read) ENABLED START #
        """Return the Tune_h attribute."""
        return self._tune_h
        # PROTECTED REGION END #    //  E2RingSimulator.Tune_h_read
    def read_Tune_v(self):
        # PROTECTED REGION ID(E2RingSimulator.Tune_v_read) ENABLED START #
        """Return the Tune_v attribute."""
        return self._tune_v
        # PROTECTED REGION END #    //  E2RingSimulator.Tune_v_read
    def read_Chromaticity_h(self):
        # PROTECTED REGION ID(E2RingSimulator.Chromaticity_h_read) ENABLED START #
        """Return the Chromaticity_h attribute."""
        return self._chromaticity_h
        # PROTECTED REGION END #    //  E2RingSimulator.Chromaticity_h_read
    def read_Chromaticity_v(self):
        # PROTECTED REGION ID(E2RingSimulator.Chromaticity_v_read) ENABLED START #
        """Return the Chromaticity_v attribute."""
        return self._chromaticity_v
        # PROTECTED REGION END #    //  E2RingSimulator.Chromaticity_v_read
    def read_Emittance_h(self):
        # PROTECTED REGION ID(E2RingSimulator.Emittance_h_read) ENABLED START #
        """Return the Emittance_h attribute."""
        return self._emittance_h
        # PROTECTED REGION END #    //  E2RingSimulator.Emittance_h_read
    def read_Emittance_v(self):
        # PROTECTED REGION ID(E2RingSimulator.Emittance_v_read) ENABLED START #
        """Return the Emittance_v attribute."""
        return self._emittance_v
        # PROTECTED REGION END #    //  E2RingSimulator.Emittance_v_read
    def read_RfFrequency(self):
        # PROTECTED REGION ID(E2RingSimulator.RfFrequency_read) ENABLED START #
        """Return the RfFrequency attribute."""
        return self._rf_frequency
        # PROTECTED REGION END #    //  E2RingSimulator.RfFrequency_read
    def write_RfFrequency(self, value):
        # PROTECTED REGION ID(E2RingSimulator.RfFrequency_write) ENABLED START #
        """Set the RfFrequency attribute."""
        with self._radlock:
            for irf in self.ind_rf:
                self.ring[irf].Frequency = value
            self._rf_frequency = value
        # PROTECTED REGION END #    //  E2RingSimulator.RfFrequency_write
    def read_RfVoltage(self):
        # PROTECTED REGION ID(E2RingSimulator.RfVoltage_read) ENABLED START #
        """Return the RfVoltage attribute."""
        return self._rf_voltage
        # PROTECTED REGION END #    //  E2RingSimulator.RfVoltage_read
    def write_RfVoltage(self, value):
        # PROTECTED REGION ID(E2RingSimulator.RfVoltage_write) ENABLED START #
        """Set the RfVoltage attribute."""
        volt = value / len(self.ind_rf)
        with self._radlock:
            for irf in self.ind_rf:
                self.ring[irf].Voltage = volt
            self._rf_voltage = value
        # PROTECTED REGION END #    //  E2RingSimulator.RfVoltage_write
    def read_Radiation(self):
        # PROTECTED REGION ID(E2RingSimulator.Radiation_read) ENABLED START #
        """Return the Radiation attribute."""
        return self._radiation
        # PROTECTED REGION END #    //  E2RingSimulator.Radiation_read
    def write_Radiation(self, value):
        # PROTECTED REGION ID(E2RingSimulator.Radiation_write) ENABLED START #
        """Set the Radiation attribute."""
        with self._radlock:
            self._radiation = value
            if self._radiation:
                self.ring.radiation_on()
            else:
                self.ring.radiation_off()
        # PROTECTED REGION END #    //  E2RingSimulator.Radiation_write
    def read_LoopTime(self):
        # PROTECTED REGION ID(E2RingSimulator.LoopTime_read) ENABLED START #
        """Return the LoopTime attribute."""
        return self._loop_time
        # PROTECTED REGION END #    //  E2RingSimulator.LoopTime_read
    def read_Counter(self):
        # PROTECTED REGION ID(E2RingSimulator.Counter_read) ENABLED START #
        """Return the Counter attribute."""
        return self._counter
        # PROTECTED REGION END #    //  E2RingSimulator.Counter_read
    def read_Mode(self):
        # PROTECTED REGION ID(E2RingSimulator.Mode_read) ENABLED START #
        """Return the Mode attribute."""
        return self._mode
        # PROTECTED REGION END #    //  E2RingSimulator.Mode_read
    def write_Mode(self, value):
        # PROTECTED REGION ID(E2RingSimulator.Mode_write) ENABLED START #
        """Set the Mode attribute."""
        self._mode = value
        # PROTECTED REGION END #    //  E2RingSimulator.Mode_write
    def read_AtxEvery(self):
        # PROTECTED REGION ID(E2RingSimulator.AtxEvery_read) ENABLED START #
        """Return the AtxEvery attribute."""
        return self._atx_every
        # PROTECTED REGION END #    //  E2RingSimulator.AtxEvery_read
    def write_AtxEvery(self, value):
        # PROTECTED REGION ID(E2RingSimulator.AtxEvery_write) ENABLED START #
        """Set the AtxEvery attribute."""
        self._atx_every = value
        pass
        # PROTECTED REGION END #    //  E2RingSimulator.AtxEvery_write
    def read_RingName(self):
        # PROTECTED REGION ID(E2RingSimulator.RingName_read) ENABLED START #
        """Return the RingName attribute."""
        return self._ring_name
        # PROTECTED REGION END #    //  E2RingSimulator.RingName_read
    def write_RingName(self, value):
        # PROTECTED REGION ID(E2RingSimulator.RingName_write) ENABLED START #
        """Set the RingName attribute."""
        pass
        # PROTECTED REGION END #    //  E2RingSimulator.RingName_write
    def is_RingName_allowed(self, attr):
        # PROTECTED REGION ID(E2RingSimulator.is_RingName_allowed) ENABLED START #
        if attr==attr.READ_REQ:
            return self.get_state() not in [DevState.FAULT,DevState.MOVING]
        else:
            return self.get_state() not in [DevState.FAULT,DevState.OFF,DevState.ON]
        # PROTECTED REGION END #    //  E2RingSimulator.is_RingName_allowed

    def read_SimulationStarted(self):
        # PROTECTED REGION ID(E2RingSimulator.SimulationStarted_read) ENABLED START #
        """Return the SimulationStarted attribute."""
        return self._simulation_started
        # PROTECTED REGION END #    //  E2RingSimulator.SimulationStarted_read
    def read_TbT_BufferSize(self):
        # PROTECTED REGION ID(E2RingSimulator.TbT_BufferSize_read) ENABLED START #
        """Return the TbT_BufferSize attribute."""
        return self._tb_t__buffer_size
        # PROTECTED REGION END #    //  E2RingSimulator.TbT_BufferSize_read
    def write_TbT_BufferSize(self, value):
        # PROTECTED REGION ID(E2RingSimulator.TbT_BufferSize_write) ENABLED START #
        """Set the TbT_BufferSize attribute."""
        self._tb_t__buffer_size = value
        # PROTECTED REGION END #    //  E2RingSimulator.TbT_BufferSize_write
    def write_ErrMessage(self, value):
        # PROTECTED REGION ID(E2RingSimulator.ErrMessage_write) ENABLED START #
        """Set the ErrMessage attribute."""
        pass
        # PROTECTED REGION END #    //  E2RingSimulator.ErrMessage_write
    def read_Energy(self):
        # PROTECTED REGION ID(E2RingSimulator.Energy_read) ENABLED START #
        """Return the Energy attribute."""
        return self._energy
        # PROTECTED REGION END #    //  E2RingSimulator.Energy_read
    def read_UpdateTime(self):
        # PROTECTED REGION ID(E2RingSimulator.UpdateTime_read) ENABLED START #
        """Return the UpdateTime attribute."""
        return self._update_time
        # PROTECTED REGION END #    //  E2RingSimulator.UpdateTime_read
    def read_ClosedOrbit_h(self):
        # PROTECTED REGION ID(E2RingSimulator.ClosedOrbit_h_read) ENABLED START #
        """Return the ClosedOrbit_h attribute."""
        return self._closed_orbit_h
        # PROTECTED REGION END #    //  E2RingSimulator.ClosedOrbit_h_read
    def read_ClosedOrbit_v(self):
        # PROTECTED REGION ID(E2RingSimulator.ClosedOrbit_v_read) ENABLED START #
        """Return the ClosedOrbit_v attribute."""
        return self._closed_orbit_v
        # PROTECTED REGION END #    //  E2RingSimulator.ClosedOrbit_v_read
    def read_Beta_h(self):
        # PROTECTED REGION ID(E2RingSimulator.Beta_h_read) ENABLED START #
        """Return the Beta_h attribute."""
        return self._beta_h
        # PROTECTED REGION END #    //  E2RingSimulator.Beta_h_read
    def read_Beta_v(self):
        # PROTECTED REGION ID(E2RingSimulator.Beta_v_read) ENABLED START #
        """Return the Beta_v attribute."""
        return self._beta_v
        # PROTECTED REGION END #    //  E2RingSimulator.Beta_v_read
    def read_Eta_h(self):
        # PROTECTED REGION ID(E2RingSimulator.Eta_h_read) ENABLED START #
        """Return the Eta_h attribute."""
        return self._eta_h
        # PROTECTED REGION END #    //  E2RingSimulator.Eta_h_read
    def read_Eta_v(self):
        # PROTECTED REGION ID(E2RingSimulator.Eta_v_read) ENABLED START #
        """Return the Eta_v attribute."""
        return self._eta_v
        # PROTECTED REGION END #    //  E2RingSimulator.Eta_v_read
    def read_TbT_InCoord(self):
        # PROTECTED REGION ID(E2RingSimulator.TbT_InCoord_read) ENABLED START #
        """Return the TbT_InCoord attribute."""
        return self._tb_t__in_coord
        # PROTECTED REGION END #    //  E2RingSimulator.TbT_InCoord_read
    def write_TbT_InCoord(self, value):
        # PROTECTED REGION ID(E2RingSimulator.TbT_InCoord_write) ENABLED START #
        """Set the TbT_InCoord attribute."""
        if len(value) != E2RingSimulator.TBT_COORD_NB:
            mess = "Wrong number of elements. You sent {} data while we need {}".format(len(value),E2RingSimulator.TBT_COORD_NB)
            tango.Except.throw_exception("E2RingSimulator_WrongArg",mess,"E2RingSimulator.write_TbT_InCoord()")

        self._tb_t__in_coord = value
        # PROTECTED REGION END #    //  E2RingSimulator.TbT_InCoord_write
    def read_Emittances_BPM_H(self):
        # PROTECTED REGION ID(E2RingSimulator.Emittances_BPM_H_read) ENABLED START #
        """Return the Emittances_BPM_H attribute."""
        return self._emittances__bpm__h
        # PROTECTED REGION END #    //  E2RingSimulator.Emittances_BPM_H_read
    def read_Emittances_BPM_V(self):
        # PROTECTED REGION ID(E2RingSimulator.Emittances_BPM_V_read) ENABLED START #
        """Return the Emittances_BPM_V attribute."""
        return self._emittances__bpm__v
        # PROTECTED REGION END #    //  E2RingSimulator.Emittances_BPM_V_read
    def read_Emittances_PinHoles_H(self):
        # PROTECTED REGION ID(E2RingSimulator.Emittances_PinHoles_H_read) ENABLED START #
        """Return the Emittances_PinHoles_H attribute."""
        return self._emittances__pin_holes__h
        # PROTECTED REGION END #    //  E2RingSimulator.Emittances_PinHoles_H_read
    def read_Emittances_PinHoles_V(self):
        # PROTECTED REGION ID(E2RingSimulator.Emittances_PinHoles_V_read) ENABLED START #
        """Return the Emittances_PinHoles_V attribute."""
        return self._emittances__pin_holes__v
        # PROTECTED REGION END #    //  E2RingSimulator.Emittances_PinHoles_V_read
    def read_S_bpms(self):
        # PROTECTED REGION ID(E2RingSimulator.S_bpms_read) ENABLED START #
        """Return the S_bpms attribute."""
        return self._s_bpms
        # PROTECTED REGION END #    //  E2RingSimulator.S_bpms_read
    def write_S_bpms(self, value):
        # PROTECTED REGION ID(E2RingSimulator.S_bpms_write) ENABLED START #
        """Set the S_bpms attribute."""
        pass
        # PROTECTED REGION END #    //  E2RingSimulator.S_bpms_write
    def read_S_pinholes(self):
        # PROTECTED REGION ID(E2RingSimulator.S_pinholes_read) ENABLED START #
        """Return the S_pinholes attribute."""
        return self._s_pinholes
        # PROTECTED REGION END #    //  E2RingSimulator.S_pinholes_read
    def write_S_pinholes(self, value):
        # PROTECTED REGION ID(E2RingSimulator.S_pinholes_write) ENABLED START #
        """Set the S_pinholes attribute."""
        pass
        # PROTECTED REGION END #    //  E2RingSimulator.S_pinholes_write
    def read_HPositionsTbT(self):
        # PROTECTED REGION ID(E2RingSimulator.HPositionsTbT_read) ENABLED START #
        """Return the HPositionsTbT attribute."""
        return self._h_positions_tb_t
        # PROTECTED REGION END #    //  E2RingSimulator.HPositionsTbT_read
    def read_VPositionsTbT(self):
        # PROTECTED REGION ID(E2RingSimulator.VPositionsTbT_read) ENABLED START #
        """Return the VPositionsTbT attribute."""
        return self._v_positions_tb_t
        # PROTECTED REGION END #    //  E2RingSimulator.VPositionsTbT_read
    # --------
    # Commands
    # --------


    @command(
    )
    @DebugIt()
    def On(self):
        # PROTECTED REGION ID(E2RingSimulator.On) ENABLED START #
        """
        Start the simulation
        :rtype: PyTango.DevVoid
        """
        if self.thread is None:
           self.thread = e2th.E2RingThread(self)
           self.thread.start()
           self.set_state(tango.DevState.ON)
        if self.update_thread is None:
           self.update_thread = e2update.E2RingUpdateThread(self)
           self.update_thread.start()
           self.set_state(tango.DevState.ON)
        # PROTECTED REGION END #    //  E2RingSimulator.On

    def is_On_allowed(self):
        # PROTECTED REGION ID(E2RingSimulator.is_On_allowed) ENABLED START #
        return self.get_state() not in [DevState.FAULT,DevState.MOVING,DevState.ON,DevState.RUNNING]
        # PROTECTED REGION END #    //  E2RingSimulator.is_On_allowed

    @command(
    )
    @DebugIt()
    def Off(self):
        # PROTECTED REGION ID(E2RingSimulator.Off) ENABLED START #
        """
        Stop the simulation
        :rtype: PyTango.DevVoid
        """
        if self.thread:
            self.thread.stop()
            self.debug_stream("joining thread")
            try:
                self.thread.join()
            except Exception as e:
                print("The exception has been Handled in the Main, Details of the Exception are:\n",e)
            self.thread = None
        if self.update_thread:
            self.update_thread.stop()
            self.debug_stream("joining update_thread")
            try:
                self.update_thread.join()
            except Exception as e:
                print("The exception has been Handled in the Main, Details of the Exception are:\n",e)
            self.update_thread = None
        self.set_state(tango.DevState.OFF)
        # PROTECTED REGION END #    //  E2RingSimulator.Off

    def is_Off_allowed(self):
        # PROTECTED REGION ID(E2RingSimulator.is_Off_allowed) ENABLED START #
        return self.get_state() not in [DevState.FAULT,DevState.MOVING,DevState.OFF,DevState.RUNNING]
        # PROTECTED REGION END #    //  E2RingSimulator.is_Off_allowed

    @command(
    )
    @DebugIt()
    def Reset(self):
        # PROTECTED REGION ID(E2RingSimulator.Reset) ENABLED START #
        """
        Reset the device after error
        :rtype: PyTango.DevVoid
        """
        pass
        # PROTECTED REGION END #    //  E2RingSimulator.Reset

    def is_Reset_allowed(self):
        # PROTECTED REGION ID(E2RingSimulator.is_Reset_allowed) ENABLED START #
        return self.get_state() not in [DevState.MOVING,DevState.OFF,DevState.ON,DevState.RUNNING]
        # PROTECTED REGION END #    //  E2RingSimulator.is_Reset_allowed

    @command(
    )
    @DebugIt()
    def ResetSimu(self):
        # PROTECTED REGION ID(E2RingSimulator.ResetSimu) ENABLED START #
        """
        Reset the whole simulation
        :rtype: PyTango.DevVoid
        """
        pass
        # PROTECTED REGION END #    //  E2RingSimulator.ResetSimu

    def is_ResetSimu_allowed(self):
        # PROTECTED REGION ID(E2RingSimulator.is_ResetSimu_allowed) ENABLED START #
        return self.get_state() not in [DevState.FAULT,DevState.MOVING,DevState.RUNNING]
        # PROTECTED REGION END #    //  E2RingSimulator.is_ResetSimu_allowed

    @command(
        dtype_in='DevLong',
        doc_in="Index in pinhole list"
               "0 -> Pinhole in cell 07"
               "1 -> Pinhole in cell 17"
               "2 -> Pinhole in cell 25"
               "3 -> Pinhole in cell 27"
               "4 -> Pinhole in cell 01",
        dtype_out='DevVarDoubleArray',
        doc_out="Pinhole data:"
                "d[0] = sigma_h power 2"
                "d[1] = sigma_v power 2"
                "d[2] = hv_coeff"
                "d[3] = Beam position_h at pinhole place"
                "d[4] = Beam position_v at pinhole place",
        display_level=DispLevel.EXPERT,
    )
    @DebugIt()
    def GetPinholeData(self, argin):
        # PROTECTED REGION ID(E2RingSimulator.GetPinholeData) ENABLED START #
        """
            Get pinhole related data for one of the 5 pinholes camera
            installed in EBS
        :param argin:     Index in pinhole list
            0 -> Pinhole in cell 07
            1 -> Pinhole in cell 17
            2 -> Pinhole in cell 25
            3 -> Pinhole in cell 27
            4 -> Pinhole in cell 01
        :type argin: PyTango.DevLong

        :return:     Pinhole data:
            d[0] = sigma_h power 2
            d[1] = sigma_v power 2
            d[2] = hv_coeff
            d[3] = Beam position_h at pinhole place
            d[4] = Beam position_v at pinhole place
        :rtype: PyTango.DevVarDoubleArray
        """
        return [0.0]
        # PROTECTED REGION END #    //  E2RingSimulator.GetPinholeData

    def is_GetPinholeData_allowed(self):
        # PROTECTED REGION ID(E2RingSimulator.is_GetPinholeData_allowed) ENABLED START #
        return self.get_state() not in [DevState.FAULT,DevState.MOVING,DevState.OFF,DevState.RUNNING]
        # PROTECTED REGION END #    //  E2RingSimulator.is_GetPinholeData_allowed

    @command(
    )
    @DebugIt()
    def ResetToDesign(self):
        # PROTECTED REGION ID(E2RingSimulator.ResetToDesign) ENABLED START #
        """
        Put all correction to 0.
        :rtype: PyTango.DevVoid
        """
        pass
        # PROTECTED REGION END #    //  E2RingSimulator.ResetToDesign

# ----------
# Run server
# ----------

# PROTECTED REGION ID(E2RingSimulator.custom_code) ENABLED START #
# PROTECTED REGION END #    //  E2RingSimulator.custom_code


def main(args=None, **kwargs):
    """Main function of the E2RingSimulator module."""
    # PROTECTED REGION ID(E2RingSimulator.main) ENABLED START #
    return run((E2RingSimulator,), args=args, **kwargs)
    # PROTECTED REGION END #    //  E2RingSimulator.main

# PROTECTED REGION ID(E2RingSimulator.custom_functions) ENABLED START #
# PROTECTED REGION END #    //  E2RingSimulator.custom_functions


if __name__ == '__main__':
    main()
